/*-
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright (c) 2018 Rafal Kozik <rafkozik@gmail.com>
 * All rights reserved.
 *
 * AssemblerThermometer
 */

/* https://wiki.hackerspace-krk.pl/poradniki/stm32/debugging */

/* Include file with registers definition */
.include "registers.S"
.global _start

/* Constans */
.set COUNTS,            (1000)
.set DELAY,             (1000)
.set DIGITS,            (4)

.set DQ,                (13)
.set LED_C,             (3)
.set LED_K,             (4)
.set BUTTON,            (8)

.set TO_KELVIN,         (2732)

.set CELIUS,            (0)
.set KELVIN,            (1)

.set BUTTON_OFF,        (0)
.set BUTTON_ON_DEB,     (1)
.set BUTTON_ON,         (2)
.set BUTTON_OFF_DEB,    (3)
.set BUTTON_TH,         (5)

/* Global variables addreses */
/* Start from begining of RAM */
.set digit_array,       (SRAM + 0x00)
.set multiplexing,      (SRAM + 0x04)
.set scale,             (SRAM + 0x08)
.set button_state,      (SRAM + 0x0c)
.set button_counter,    (SRAM + 0x10)

/* Vector table start */
.long   (SRAM + SRAM_SIZE)  /* Stack pointer - end of RAM*/
.long   (_start+1)              /* Reset vector +1 - THUMB*/
.long   (hard_fault+1)          /* Non maskable interrupt */
.long   (hard_fault+1)

/* Hard fault */
/* Stay in infinite loop... */
hard_fault:
B       hard_fault

/* Main program */
_start:
/* Configuration */
/* Enble GPIOA, B and C clock */
LDR     R0, =RCC
LDR     R1, =(RCC_AHBENR_IOPAEN|RCC_AHBENR_IOPBEN|RCC_AHBENR_IOPCEN)
STR     R1, [R0, #RCC_AHBENR]

/* GPIOA */
/* 0 to 11 as out, 13,14 as AF (SWDIO, SWCLK) */
LDR     R0, =GPIOA
LDR     R1, =((1<<0)|(1<<2)|(1<<4)|(1<<6)|(1<<8)|(1<<10)|(1<<12)|(1<<14)|(1<<16)|(1<<18)|(1<<20)|(1<<22)|(2<<26)|(2<<28))
STR     R1, [R0, #GPIOx_MODER]

/* GPIOB */
/* 3,4 as out */
LDR     R0, =GPIOB
LDR     R1, =((1<<6)|(1<<8))
STR     R1, [R0, #GPIOx_MODER]
/* 8 wit pull-up */
LDR     R1, =(1<<16)
STR     R1, [R0, #GPIOx_PUPDR]

/* GPIOC */
LDR     R0, =GPIOC
/* PC13 as out */
LDR     R1, =(1<<(DQ*2))
STR     R1, [R0, #GPIOx_MODER]
/* Open drain */
LDR     R1, =(1<<DQ)
STR     R1, [R0, #GPIOx_OTYPER]

/* Initialize variables with 0 */
MOVS    R5, #CELIUS
LDR     R3, =scale
STR     R5, [R3, #0]
MOVS    R5, #0
LDR     R3, =digit_array
STR     R5, [R3, #0]
LDR     R3, =multiplexing
STR     R5, [R3, #0]
LDR     R3, =button_state
STR     R5, [R3, #0]
LDR     R3, =button_counter
STR     R5, [R3, #0]

LDR     R2, =COUNTS

/* Main loop */
loop:
SUBS    R2, #1
BNE     main_loop_display

LDR     R2, =COUNTS

LDR     R3, =scale
LDR     R3, [R3, #0]
CMP     R3, #CELIUS
BNE     use_kelvin
MOVS    R1, (1<<LED_C)
B       set_LEDs
use_kelvin:
MOVS    R1, (1<<LED_K)
set_LEDs:
LDR     R0, =GPIOB
STR     R1, [R0, #GPIOx_ODR]

CMP     R5, #0
BNE     init
MOVS    R5, #1
BL      get_tmp
B       main_loop_display
init:
MOVS    R5, #0
BL      DS18B20_init_conv
main_loop_display:
/* Call display_number function */
BL      display_number
BL      button

/* Delay */
LDR     R0, =DELAY
BL      delay
B       loop

/* Functions */

/* Delay function */
/* Delay R0 iterations */
/* R0/2+1 us */
delay:
PUSH    {R0}
delay_loop:
SUBS    R0, #1
BNE     delay_loop
POP     {R0}
BX      LR

/* Function that perform multiplexing */
/* Digits stored in global array digit_array */
/* Number of current display stored in global variable multiplexing */
display_number:
PUSH    {R0,R1,R2,R3,R4}        /* Backup registers on stack */
LDR     R1, =multiplexing       /* Load address of global variable */
LDR     R0, [R1, #0]            /* Load variable from SRAM to R0 */
LDR     R1, =digit_array        /* Load address if global array */
LDRB    R1, [R1, R0]            /* Load byte from address R1, with offset R0 */
LDR     R2, =numbers            /* Load address of const array */
LDRB    R1, [R2, R1]            /* Load byte with offset R1 from address R2 */
CMP     R0, #2                  /* Dot point on display 2 */
BNE     not_dp                  /* if (R0 != 2) goto not_dp */
MOVS    R4, #DOT_POINT          /* Rk = DOT_POINT */
ANDS    R1, R4                  /* R1 |= R4 */
not_dp:
LDR     R2, =digit              /* Load address of const array */
LDRB    R2, [R2, R0]            /* Load byte from address R2 with offset R0 */
LSLS    R2, #8                  /* Shift left R2 */
ORRS    R1, R2                  /* R1 |= R2 */
LDR     R2, =GPIOA              /* Load Address of GPIOA register */
STR     R1, [R2, #GPIOx_ODR]    /* Store R1 to GPIOA_ODR */
ADDS    R0, #1                  /* R0++ */
CMP     R0, #DIGITS             /* Compare R0 with constant */
BNE     dn_ne                   /* If (R0 != DIGITS) goto dn_ne; */ 
MOVS    R0, #0                  /* R0 = 0 (if (R0 == DIGITS)) */
dn_ne:
LDR     R1, =multiplexing       /* Load address of global variable */
STR     R0, [R1, #0]            /* Stor R0 in SRAM */
POP     {R0,R1,R2,R3,R4}        /* Restore registers from stack*/
BX      LR                      /* Return */

/* Const array of numbers code */
.set DOT_POINT,         (0xfb)
numbers:
.byte 0x05, 0x7d, 0x46, 0x54, 0x3c
.byte 0x94, 0x84, 0x5d, 0x04, 0x14
.byte 0xfe, 0x00

/* Const array that convert digit number to active pin */
digit:
.byte 0x04, 0x02, 0x01, 0x08

/* Convert binary R0 to BCD amd store in global variable */
/* http://www.piclist.com/techref/language/ccpp/convertbase.htm */
to_bcd:
PUSH    {R0,R1,R2,R3,R4,R5,R6,R7}   /* Store registers on stack */
MOVS    R7, #0              /* flag that stor if on last screen should be - */
CMP     R0, #0              /* R0 >= 0 */
BGE     to_bcd_convert      /* if (R0 >= 0) goto to_bcd_convert; */
MOVS    R7, #1              /* set flag */
LDR     R1, =10             /* - is char number 10 */
LDR     R2, =digit_array    /* get address of global array */
STRB    R1, [R2, #0]        /* stor R1 as digit_array[0] */
RSBS    R0, R0,  #0         /* R0 = -R0 */
to_bcd_convert:
MOVS    R5, 0xF
/* R2 = (R0>>4)  & 0xF */
LSRS    R2, R0, #4
ANDS    R2, R5
/* R3 = (R0>>8)  & 0xF */
LSRS    R3, R0, #8
ANDS    R3, R5
/* R4 = (R0>>12) & 0xF */
LSRS    R4, R0, #12
ANDS    R4, R5

/* R1 = 6*(R4 + R3 + R2) + (R0 & 0xF) */
ADDS    R1, R2, R3
ADDS    R1, R4
MOVS    R6, #6
MULS    R1, R1, R6
ANDS    R0, R5
ADDS    R1, R0

/* R6 = (d0 * 0xCD) >> 11 */
MOVS    R6, 0xCD
MULS    R6, R1, R6
LSRS    R6, #11
/* R0 = R0 - 10*R6 */
MOVS    R5, #10
MULS    R5, R5, R6
SUBS    R1, R5

/* Store R1 to global array */
LDR     R5, =digit_array
STRB    R1, [R5, #3]

/* R2 = R6 + 9*R4 + 5*R3 + R2 */
ADDS    R2, R6
MOVS    R1, #9
MULS    R1, R1, R4
ADDS    R2, R1
MOVS    R1, #5
MULS    R1, R1, R3
ADDS    R2, R1
/* R5 = (R2 * 0xCD) >> 11 */
MOVS    R6, 0xCD
MULS    R6, R6, R2
LSRS    R6, #11
/* R2 = R2 - 10*R5 */
MOVS    R5, #10
MULS    R5, R5, R6
SUBS    R2, R5

/* Store R2 to global array */
LDR     R5, =digit_array
STRB    R2, [R5, #2]

/* R3 = R5 + 2*R3 */
LSLS    R3, #1
ADDS    R3, R6
/* R5 = (R3 * 0x1A) >> 8 */
MOVS    R6, #0x1A
MULS    R6, R6, R3
LSRS    R6, #8
/* R3 = R3 - 10*R5 */
MOVS    R5, #10
MULS    R5, R5, R6
SUBS    R3, R5

/* Store R3 to global array */
LDR     R5, =digit_array
STRB    R3, [R5, #1]

CMP     R7, #1              /* check flag */
BEQ     to_bcd_end          /* if flag set, show '-' and go to end */

/* R4 = R5 + 4*R4 */
LSLS    R4, #2
ADDS    R4, R6
/* R5 = (R4 * 0x1A) >> 8 */
MOVS    R6, #0x1A
MULS    R6, R6, R4
LSRS    R6, #8
/* R4 = R4 - 10*R5 */
MOVS    R5, #10
MULS    R5, R5, R6
SUBS    R4, R5

/* Store R4 to global array */
LDR     R5, =digit_array
STRB    R4, [R5, #0]

to_bcd_end:
POP     {R0,R1,R2,R3,R4,R5,R6,R7}   /* Restore registers */
BX      LR


/* DS18B20 commands */
.set DS18B20_skip_rom,      (0xCC)
.set DS18B20_convert_T,     (0x44)
.set DS18B20_read_sp,       (0xBE)

get_tmp:
PUSH    {R0,R1,R2,R3,R4,R5,R6,R7,LR}   /* Restore registers */
LDR     R1, =GPIOC

get_tmp_work:
/* reset transmision */
BL      DS18B20_reset

/* init data read */
MOVS    R3, #DS18B20_skip_rom
BL      DS18B20_send
MOVS    R3, #DS18B20_read_sp
BL      DS18B20_send

/* read temperature */
BL      DS18B20_receive
MOVS    R6, R3
BL      DS18B20_receive
LSLS    R3, #8
ORRS    R6, R3
SXTH    R6, R6

MOVS    R3, #10
MULS    R6, R3
ASRS    R6, #4

/* Convert to Kelvin if needed */
LDR     R3, =scale
LDR     R3, [R3, #0]
CMP     R3, #CELIUS
BEQ     change_to_bcd
LDR     R3, =TO_KELVIN
ADDS    R6, R3

change_to_bcd:
MOVS    R0, R6
BL      to_bcd

POP     {R0,R1,R2,R3,R4,R5,R6,R7,PC}   /* Restore registers */

DS18B20_init_conv:
/* init new conversion */
PUSH    {R0,R1,R2,R3,R4,R5,R6,R7,LR}
LDR     R1, =GPIOC
BL      DS18B20_reset
MOVS    R3, #DS18B20_skip_rom
BL      DS18B20_send
MOVS    R3, #DS18B20_convert_T
BL      DS18B20_send
POP     {R0,R1,R2,R3,R4,R5,R6,R7,PC}


DS18B20_reset:
PUSH    {LR}
MOVS    R2, #0
STR     R2, [R1, #GPIOx_ODR]

LDR     R0, =1000
BL      delay

LDR     R2, =(1<<DQ)
STR     R2, [R1, #GPIOx_ODR]

LDR     R0, =1000
BL      delay

POP     {PC}

/* send byte from R3 */
DS18B20_send:
PUSH    {LR}
MOVS    R4, #8

DS18B20_send_loop:
LSRS    R3, #1                  /* Move LSB to C flag */
BHS     DS18B20_send_loop_one   /* if (C == 1) goto */
BL      DS18B20_write_zero      
B       DS18B20_send_loop_end
DS18B20_send_loop_one:
BL      DS18B20_write_one
DS18B20_send_loop_end:
SUBS    R4, #1
BNE     DS18B20_send_loop

POP     {PC}

DS18B20_write_one:
PUSH    {LR}

MOVS    R2, #0
STR     R2, [R1, #GPIOx_ODR]

MOVS    R0, #16                 /* ~10 us */
BL      delay

LDR     R2, =(1<<DQ)
STR     R2, [R1, #GPIOx_ODR]

MOVS    R0, #96                 /* ~50 us */
BL      delay

POP     {PC}

DS18B20_write_zero:
PUSH    {LR}

MOVS    R2, #0
STR     R2, [R1, #GPIOx_ODR]

MOVS    R0, #116                 /* ~60 us */
BL      delay

LDR     R2, =(1<<DQ)
STR     R2, [R1, #GPIOx_ODR]

MOVS    R0, #6                 /* ~5 us */
BL      delay

POP     {PC}

/* result stored in R3 */
DS18B20_receive:
PUSH    {LR}
MOVS    R3, #0
MOVS    R4, #8
MOVS    R5, #(1<<7)

DS18B20_receive_loop:
LSRS    R3, #1

MOVS    R2, #0
STR     R2, [R1, #GPIOx_ODR]
MOVS    R0, #1                  /* ~2 us */
BL      delay

LDR     R2, =(1<<DQ)
STR     R2, [R1, #GPIOx_ODR]
MOVS    R0, #16                 /* ~10 us */
BL      delay

LDR     R2, [R1, #GPIOx_IDR]
LSRS    R2, #(DQ-7)
ANDS    R2, R5
ORRS    R3, R2

MOVS    R0, #92                 /* ~48 us */
BL      delay

SUBS    R4, #1
BNE     DS18B20_receive_loop

POP     {PC}

button:
PUSH    {R0, R1, R2, R3, R4, R5, R6, R7}

LDR     R2, =GPIOB
LDR     R2, [R2, #GPIOx_IDR]
LSRS    R2, #BUTTON
MOVS    R1, #1
ANDS    R2, R1

LDR     R0, =button_state
LDR     R1, [R0, #0]

CMP     R1, #BUTTON_OFF
BEQ     button_off
CMP     R1, #BUTTON_ON_DEB
BEQ     button_on_deb
CMP     R1, #BUTTON_ON
BEQ     button_on
CMP     R1, #BUTTON_ON_DEB
BEQ     button_off_deb
B       button_default

button_off:
CMP     R2, #1
BEQ     button_end      /* not pushed */
MOVS    R1, #BUTTON_ON_DEB
LDR     R2, =button_counter
MOVS    R3, #0
STR     R3, [R2, #0]
B button_end

button_on_deb:
CMP     R2, #1
BEQ     button_released
LDR     R2, =button_counter
LDR     R3, [R2, #0]
ADDS    R3, #1
STR     R3, [R2, #0]
CMP     R3, #BUTTON_TH
BNE     button_end
MOVS    R1, BUTTON_ON
LDR     R4, =scale
LDR     R5, [R4, #0]
MOVS    R6, #1
SUBS    R6, R5
STR     R6, [R4, #0]
B button_end
button_released:
MOVS    R1, #BUTTON_OFF
B button_end

button_on:
CMP     R2, #1
BNE     button_end
MOVS    R1, BUTTON_OFF_DEB
LDR     R2, =button_counter
MOVS    R3, #0
STR     R3, [R2, #0]
B button_end

button_off_deb:
CMP     R2, #1
BNE     button_not_released
LDR     R2, =button_counter
LDR     R3, [R2, #0]
ADDS    R3, #1
STR     R3, [R2, #0]
CMP     R3, #BUTTON_TH
BNE     button_end
MOVS    R1, BUTTON_OFF
button_not_released:
MOVS    R1, BUTTON_ON
B button_end

button_default:
MOVS    R1, #BUTTON_OFF

button_end:
STR     R1, [R0, #0]
POP     {R0, R1, R2, R3, R4, R5, R6, R7}
BX      LR
